import os
import re
import sys
from collections import defaultdict
from logging import getLogger
from typing import DefaultDict, Literal

import yaml
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    ValidationError,
    ValidationInfo,
    field_validator,
)

logger = getLogger("uvicorn.error")


class General(BaseModel):
    prefix: str = "junos"
    ssh_config: str | None = None

    @field_validator("ssh_config", mode="after")
    @classmethod
    def check_exist_file(cls, path: str) -> str:
        abs_path = os.path.abspath(os.path.expanduser(path))
        if not os.path.isfile(abs_path):
            raise ValueError(f"file({abs_path}) does not exist")
        return abs_path


class Credential(BaseModel):
    username: str
    password: str


class Module(BaseModel):
    tables: list[str]

    @field_validator("tables", mode="before")
    @classmethod
    def check_exist_optables(cls, tables: list[str], info: ValidationInfo) -> list[str]:
        if isinstance(info.context, dict):
            optables = info.context.get("optables", dict())
            for table in tables:
                if table not in optables:
                    raise ValueError(f"table({table}) does not contain optables")
        return tables


class Label(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)

    name: str
    value: str
    regex: re.Pattern | None = None

    @field_validator("regex", mode="before")
    @classmethod
    def to_re_pattern(cls, regex: str) -> re.Pattern:
        if not isinstance(regex, str):
            raise ValueError(f"regex({regex}) is not a str")
        return re.compile(regex)


class Metric(BaseModel):
    model_config = ConfigDict(coerce_numbers_to_str=True)

    name: str
    value: str
    type_: Literal["untyped", "counter", "gauge"] = Field("untyped", alias="type")
    help_: str = Field("", alias="help")
    regex: re.Pattern | None = None
    value_transform: DefaultDict[str | bool, float] | None = None
    to_unixtime: bool = False

    @field_validator("regex", mode="before")
    @classmethod
    def to_re_pattern(cls, regex: str) -> re.Pattern:
        if not isinstance(regex, str):
            raise ValueError(f"regex({regex}) is not a str")
        return re.compile(regex)

    @field_validator("value_transform", mode="before")
    @classmethod
    def to_defaultdict(cls, value_transform: dict) -> dict:
        if default := value_transform.get("_"):
            return defaultdict(lambda: float(default), value_transform)
        return defaultdict(lambda: "NaN", value_transform)


class OpTable(BaseModel):
    metrics: list[Metric]
    labels: list[Label]


class Config:
    def __init__(self) -> None:
        config = {}

        config_location = [
            "config.yml",
            os.path.expanduser("~/.junos-exporter/config.yml"),
        ]
        for c in config_location:
            if os.path.isfile(c):
                try:
                    with open(c, "r") as f:
                        config = yaml.safe_load(f)
                except ValidationError as e:
                    sys.exit(f"failed to load config file.\n{e}")

        if not config:
            sys.exit(
                "config file(./config.yml or ~/.junos-exporter/config.yml) is not found."
            )

        self.general = General(**config["general"])
        self.credentials = {
            name: Credential.model_validate(
                credential, context={"optables": config["optables"]}
            )
            for name, credential in config["credentials"].items()
        }
        self.modules = {
            name: Module.model_validate(
                module, context={"optables": config["optables"]}
            )
            for name, module in config["modules"].items()
        }
        self.optables = {
            name: OpTable(**optable) for name, optable in config["optables"].items()
        }

    @property
    def prefix(self) -> str:
        return self.general.prefix

    @property
    def ssh_config(self) -> str | None:
        return self.general.ssh_config



import argparse

import uvicorn


def cli() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l",
        "--log",
        type=str,
        default="info",
        help="logging level[default: info]",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=9326,
        help="listening port[default: 9326]",
    )
    parser.add_argument(
        "-w",
        "--workers",
        type=int,
        default=1,
        help="number of worker processes[default: 1]",
    )

    args = parser.parse_args()
    uvicorn.run(
        "junos_exporter.api:app",
        host="0.0.0.0",
        port=args.port,
        workers=args.workers,
        log_config="log_config.yml",
        log_level=args.log,
    )


if __name__ == "__main__":
    cli()



import re
from datetime import datetime, timedelta

from fastapi import HTTPException, status

from .config import Config, Label, Metric, logger
from .connector import Connector


class MetricConverter:
    def __init__(
        self,
        metric: Metric,
        labels: list[Label],
        prefix: str,
        unixtime_regex: dict[str, re.Pattern],
    ):
        if metric.type_ == "counter":
            self.name = f"{prefix}_{metric.name}_total"
        else:
            self.name = f"{prefix}_{metric.name}"
        self.value_name = metric.value
        self.type_ = metric.type_
        self.help_ = metric.help_
        self.regex = metric.regex
        self.value_transform = metric.value_transform
        self.to_unixtime = metric.to_unixtime
        self.labels = labels
        self.unixtime_regex = unixtime_regex

    def _convert_to_unixtime(self, value: str) -> float:
        if result := self.unixtime_regex["timestamp"].search(value):
            return float(
                datetime.strptime(result.group(1), "%Y-%m-%d %H:%M:%S").timestamp()
                * 1000
            )

        init_time = datetime.fromtimestamp(0)
        if result := self.unixtime_regex["wd_uptime"].search(value):
            return float(
                (
                    init_time
                    + timedelta(
                        weeks=int(result.group(1)),
                        days=int(result.group(2)),
                        hours=int(result.group(3)),
                        minutes=int(result.group(4)),
                        seconds=int(result.group(5)),
                    )
                ).timestamp()
                * 1000
            )
        elif result := self.unixtime_regex["d_uptime"].search(value):
            return float(
                (
                    init_time
                    + timedelta(
                        days=int(result.group(1)),
                        hours=int(result.group(2)),
                        minutes=int(result.group(3)),
                        seconds=int(result.group(4)),
                    )
                ).timestamp()
                * 1000
            )
        elif result := self.unixtime_regex["uptime"].search(value):
            return float(
                (
                    init_time
                    + timedelta(
                        hours=int(result.group(1)),
                        minutes=int(result.group(2)),
                        seconds=int(result.group(3)),
                    )
                ).timestamp()
                * 1000
            )
        else:
            return float(0)

    def _convert_label(self, item: dict) -> list[str]:
        label_exposition = []
        for label in self.labels:
            if label.value not in item:
                continue

            if item[label.value] is None:
                continue

            if not label.regex:
                label_exposition.append(f'{label.name}="{item[label.value]}"')
                continue

            match = label.regex.match(item[label.value])
            if match is None:
                continue
            else:
                try:
                    label_exposition.append(f'{label.name}="{match.group(1)}"')
                except IndexError:
                    continue
        return label_exposition

    def convert(self, items: list[dict]) -> str:
        exposition = []
        exposition.append(f"# HELP {self.name} {self.help_}\n")
        exposition.append(f"# TYPE {self.name} {self.type_}\n")

        for item in items:
            label_exposition = ",".join(self._convert_label(item))
            if self.value_name not in item:
                try:
                    # static value
                    exposition.append(
                        f"{self.name}{{{label_exposition}}} {float(self.value_name)}\n"
                    )
                    continue
                except (ValueError, TypeError):
                    exposition.append(f"{self.name}{{{label_exposition}}} NaN\n")
                    logger.warning(
                        f"Could not convert metric value(name: {self.name}, value(static): {self.value_name}, error: could not convert type to float)"
                    )
                    continue

            value = item[self.value_name]
            if self.regex is not None:
                match = self.regex.match(value)
                if match is None:
                    exposition.append(f"{self.name}{{{label_exposition}}} NaN\n")
                    logger.warning(
                        f"Could not convert metric value(name: {self.name}, value({self.value_name}): {value}, regex: {self.regex}, error: could not match regex)"
                    )
                    continue
                else:
                    try:
                        value = match.group(1)
                    except IndexError:
                        value = match.group()

            if self.value_transform:
                exposition.append(
                    f"{self.name}{{{label_exposition}}} {self.value_transform[value]}\n"
                )
            elif self.to_unixtime:
                try:
                    exposition.append(
                        f"{self.name}{{{label_exposition}}} {self._convert_to_unixtime(value)}\n"
                    )
                except (ValueError, TypeError):
                    exposition.append(f"{self.name}{{{label_exposition}}} NaN\n")
                    logger.warning(
                        f"Could not convert metric value(name: {self.name}, value({self.value_name}): {value}, error: could not convert to unixtime)"
                    )
            else:
                try:
                    exposition.append(
                        f"{self.name}{{{label_exposition}}} {float(value)}\n"
                    )
                except (ValueError, TypeError):
                    exposition.append(f"{self.name}{{{label_exposition}}} NaN\n")
                    logger.warning(
                        f"Could not convert metric value(metric: {self.name}, value({self.value_name}): {value}, error: could not convert type to float)"
                    )
        return "".join(exposition)


class Exporter:
    def __init__(self, converter: dict[str, list[MetricConverter]]) -> None:
        self.converter = converter

    def collect(self, connector: Connector) -> str:
        exposition: list[str] = []
        for name, metrics in self.converter.items():
            items = connector.collect(name)
            logger.debug(
                f"Start to convert table items(target: {connector.host}), table: {name})"
            )
            exposition.append("\n".join([metric.convert(items) for metric in metrics]))
            logger.debug(
                f"Completed to convert table items(target: {connector.host}), table: {name})"
            )
        return "\n".join(exposition)


class ExporterBuilder:
    def __init__(self, config: Config) -> None:
        self.converters = {}
        unixtime_regex: dict[str, re.Pattern] = {
            "timestamp": re.compile(r".*(\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d).*"),
            "wd_uptime": re.compile(r".*(\d+)w(\d+)d (\d\d):(\d\d):(\d\d).*"),
            "d_uptime": re.compile(r".*(\d+)d (\d\d):(\d\d):(\d\d).*"),
            "uptime": re.compile(r".*(\d\d):(\d\d):(\d\d).*"),
        }

        for name, module in config.modules.items():
            converter = {}
            for table in module.tables:
                converter[table] = [
                    MetricConverter(
                        metric,
                        labels=config.optables[table].labels,
                        prefix=config.prefix,
                        unixtime_regex=unixtime_regex,
                    )
                    for metric in config.optables[table].metrics
                ]
            self.converters[name] = converter

    def build(self, module_name: str) -> Exporter:
        if module_name not in self.converters:
            logger.error(f"Module is not defined(module: {module_name})")
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Module is not defined(module: {module_name})",
            )
        return Exporter(self.converters[module_name])



from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import Generator

from fastapi import Depends, FastAPI, Response
from fastapi.responses import PlainTextResponse
from starlette.exceptions import HTTPException as StarletteHTTPException

from .config import Config
from .connector import ConnecterBuilder, Connector
from .exporter import Exporter, ExporterBuilder

config = Config()


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    app.exporter = ExporterBuilder(config)  # type: ignore
    app.connector = ConnecterBuilder(config)  # type: ignore
    yield


app = FastAPI(title="junos-exporter", lifespan=lifespan)


@app.exception_handler(StarletteHTTPException)
def http_exception_handler(request, exc) -> PlainTextResponse:
    return PlainTextResponse(content=str(exc.detail), status_code=exc.status_code)


def get_connector(target: str, credential: str) -> Generator[Connector, None, None]:
    with app.connector.build(target, credential) as connector:  # type: ignore
        yield connector


@app.get("/metrics", tags=["exporter"], response_class=PlainTextResponse)
def metrics(module: str, connector: Connector = Depends(get_connector)) -> str:
    exporter: Exporter = app.exporter.build(module)  # type: ignore
    return exporter.collect(connector)


@app.get("/debug", tags=["debug"])
def debug(optable: str, connector: Connector = Depends(get_connector)) -> Response:
    return Response(content=connector.debug(optable), media_type="application/json")



import os
import re
from glob import glob

import yaml
from fastapi import HTTPException, status
from jnpr.junos import Device
from jnpr.junos.exception import ConnectError
from jnpr.junos.factory import loadyaml
from jnpr.junos.factory.cmdtable import CMDTable
from jnpr.junos.factory.optable import OpTable

# predefined table
from jnpr.junos.op.arp import ArpTable
from jnpr.junos.op.bfd import BfdSessionTable
from jnpr.junos.op.bgp import bgpTable
from jnpr.junos.op.ccc import CCCTable
from jnpr.junos.op.ddos import DDOSTable
from jnpr.junos.op.elsethernetswitchingtable import ElsEthernetSwitchingTable
from jnpr.junos.op.ethernetswitchingtable import EthernetSwitchingTable
from jnpr.junos.op.ethport import EthPortTable
from jnpr.junos.op.fpc import FpcHwTable, FpcInfoTable, FpcMiReHwTable, FpcMiReInfoTable
from jnpr.junos.op.idpattacks import IDPAttackTable
from jnpr.junos.op.intopticdiag import PhyPortDiagTable
from jnpr.junos.op.inventory import ModuleTable
from jnpr.junos.op.isis import IsisAdjacencyTable
from jnpr.junos.op.l2circuit import L2CircuitConnectionTable
from jnpr.junos.op.lacp import LacpPortTable
from jnpr.junos.op.ldp import LdpNeighborTable
from jnpr.junos.op.lldp import LLDPNeighborTable
from jnpr.junos.op.nd import NdTable
from jnpr.junos.op.ospf import (
    OspfInterfaceTable,
    OSPFIOStatsTable,
    OspfNeighborTable,
    OspfRoutesTable,
    OSPFStatsTable,
    ospfTable,
)
from jnpr.junos.op.pfestats import PFEStatsTrafficTable
from jnpr.junos.op.phyport import PhyPortErrorTable, PhyPortStatsTable, PhyPortTable
from jnpr.junos.op.ppm import PPMTable
from jnpr.junos.op.routes import RouteSummaryTable, RouteTable
from jnpr.junos.op.securityzone import SecurityZoneTable

# from jnpr.junos.op.systemstorage import SystemStorageTable
from jnpr.junos.op.taskmemory import TaskMemoryTable
from jnpr.junos.op.teddb import TedSummaryTable, TedTable
from jnpr.junos.op.vlan import VlanTable
from jnpr.junos.op.xcvr import XcvrTable

from .config import Config, Credential, logger


class Connector:
    def __init__(
        self,
        host: str,
        credential: Credential,
        textfsm_dir: str | None,
        ssh_config: str | None,
    ) -> None:
        self.device: Device = None
        self.host: str = host
        self.username: str = credential.username
        self.password: str = credential.password
        self.textfsm_dir: str | None = textfsm_dir
        self.ssh_config: str | None = ssh_config

    def __enter__(self) -> "Connector":
        try:
            logger.debug(f"Start to open netconf connection(target: {self.host})")
            if self.ssh_config is not None:
                self.device = Device(
                    host=self.host,
                    user=self.username,
                    password=self.password,
                    ssh_config=self.ssh_config,
                ).open()
            else:
                self.device = Device(
                    host=self.host,
                    user=self.username,
                    password=self.password,
                ).open()
            logger.debug(f"Completed to open netconf connection(target: {self.host})")
        except ConnectError as err:
            logger.error(
                f"Could not open netconf connection(target: {self.host}, error: {err})"
            )
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Could not open netconf connection(target: {self.host}, error: {err})",
            )
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.device.close()
        logger.debug(f"Closed netconf connection(target: {self.host})")

    def _get(self, name: str) -> CMDTable | OpTable | None:
        if issubclass(globals()[name], CMDTable):
            if self.textfsm_dir is None:
                table = globals()[name](self.device)
            else:
                table = globals()[name](self.device, template_dir=self.textfsm_dir)
        elif issubclass(globals()[name], OpTable):
            table = globals()[name](self.device)
        else:
            raise NotImplementedError

        try:
            logger.debug(
                f"Start to get table items(target: {self.host}, table: {name})"
            )
            table.get()
            logger.debug(
                f"Completed to get table items(target: {self.host}, table: {name})"
            )
            return table
        except AttributeError:
            # https://github.com/Juniper/py-junos-eznc/issues/1366
            logger.debug(
                f"Could not get table items(target: {self.host}, table: {name}, error: junos output is empty"
            )
            return None
        except Exception as err:
            logger.error(
                f"Could not get table items(target: {self.host}, table: {name}, error: {err})"
            )
            return None

    def collect(self, name: str) -> list[dict]:
        table = self._get(name)
        if table is None:
            return []

        items = []
        if isinstance(table, CMDTable):
            for t in table:
                key, table = t
                item = {}
                if type(key) is tuple:
                    for i, n in enumerate(key):
                        item[f"key.{i}"] = n
                        item[f"name.{i}"] = n
                else:
                    item["key"] = key
                    item["name"] = key

                for k, v in table.items():
                    item[k] = v
                items.append(item)
        elif isinstance(table, OpTable):
            for t in table:
                item = {}
                try:
                    if type(t.key) is tuple:
                        for i, n in enumerate(t.key):
                            item[f"key.{i}"] = n
                            item[f"name.{i}"] = n
                    else:
                        item["key"] = t.key
                        item["name"] = t.key
                except ValueError:
                    # key is not defined
                    pass

                for k, v in t.items():
                    item[k] = v
                items.append(item)
        else:
            raise NotImplementedError

        return items

    def debug(self, name: str) -> list[dict]:
        table = self._get(name)
        if table is None:
            return []
        return table.to_json()


class ConnecterBuilder:
    def __init__(self, config: Config) -> None:
        self.optabels_dir: str | None = None
        if os.path.isdir(os.path.expanduser("~/.junos-exporter/op")):
            self.optables_dir = os.path.expanduser("~/.junos-exporter/op")
        elif os.path.isdir("./op"):
            self.optables_dir = "./op"

        self.textfsm_dir: str | None = None
        if os.path.isdir(os.path.expanduser("~/.junos-exporter/textfsm")):
            self.textfsm_dir = os.path.abspath(
                os.path.expanduser("~/.junos-exporter/textfsm")
            )
        elif os.path.isdir("./textfsm"):
            self.textfsm_dir = os.path.abspath("./textfsm")

        self.credentials: dict[str, Credential] = config.credentials
        self.ssh_config: str | None = config.ssh_config
        self._load_optables()

    def _load_optables(self) -> None:
        if self.optables_dir is None:
            return
        for yml in glob(f"{self.optables_dir}/*"):
            if re.match(r".+\.(yml|yaml)$", yml):
                globals().update(loadyaml(yaml.safe_load(yml)))

    def build(self, host: str, credential_name: str) -> Connector:
        if credential_name not in self.credentials:
            logger.error(
                f"Could not build Connector(target: {host}, credential: {credential_name}, error: credential is not defined)"
            )
            raise HTTPException(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail=f"Could not build Connector(target: {host}, credential: {credential_name}, error: credential is not defined)",
            )
        return Connector(
            host=host,
            credential=self.credentials[credential_name],
            textfsm_dir=self.textfsm_dir,
            ssh_config=self.ssh_config,
        )
